@@ -55,8 +55,6 @@
 #include "parse.h"
 #include "msg.h"
 
-#define MAX_CONFIG_FILE_LINE_LEN 256
-
 
 typedef struct {
     ParsedAttribute a;
@@ -67,6 +65,7 @@ typedef struct {
 
 static ParsedAttributeWrapper *parse_config_file(char *buf,
                                                  const char *file,
+                                                 const int length,
                                                  ConfigProperties *);
 
 static int process_config_file_attributes(const char *file,
@@ -82,8 +81,6 @@ static float get_color_value(int attr,
 static int parse_config_property(const char *file, const char *line,
                                  ConfigProperties *conf);
 
-static void init_config_properties(ConfigProperties *conf);
-
 static void write_config_properties(FILE *stream, ConfigProperties *conf,
                                     char *locale);
 
@@ -99,6 +96,9 @@ static void write_config_properties(FILE *stream, ConfigProperties *conf,
  * message is printed to stderr, NV_FALSE is returned, and nothing is
  * sent to the X server.
  *
+ * NOTE: The conf->locale should have already been setup by calling
+ *       init_config_properties() prior to calling this function.
+ *
  * XXX should we do any sort of versioning to handle compatibility
  * problems in the future?
  */
@@ -106,15 +106,11 @@ static void write_config_properties(FILE *stream, ConfigProperties *conf,
 int nv_read_config_file(const char *file, const char *display_name,
                         ParsedAttribute *p, ConfigProperties *conf)
 {
-    int fd, ret;
+    int fd, ret, length;
     struct stat stat_buf;
     char *buf, *locale;
     ParsedAttributeWrapper *w = NULL;
     
-    /* initialize the ConfigProperties */
-
-    init_config_properties(conf);
-
     /* open the file */
 
     fd = open(file, O_RDONLY);
@@ -141,9 +137,11 @@ int nv_read_config_file(const char *file, const char *display_name,
         return NV_TRUE;
     }
 
+    length = stat_buf.st_size;
+
     /* map the file into memory */
 
-    buf = mmap(0, stat_buf.st_size, PROT_READ, MAP_SHARED, fd, 0);
+    buf = mmap(0, length, PROT_READ, MAP_SHARED, fd, 0);
     if (buf == (void *) -1) {
         nv_error_msg("Unable to mmap file '%s' for reading (%s).",
                      file, strerror(errno));
@@ -158,7 +156,7 @@ int nv_read_config_file(const char *file, const char *display_name,
 
     locale = strdup(conf->locale);
 
-    w = parse_config_file(buf, file, conf);
+    w = parse_config_file(buf, file, length, conf);
 
     setlocale(LC_NUMERIC, locale);
     free(locale);
@@ -215,6 +213,7 @@ int nv_write_config_file(const char *filename, CtrlHandles *h,
     ReturnStatus status;
     NVCTRLAttributeValidValuesRec valid;
     uint32 mask;
+    CtrlHandleTarget *t;
     char *tmp_d_str, *tmp, *prefix, scratch[4];
     char *locale = "C";
 
@@ -260,12 +259,19 @@ int nv_write_config_file(const char *filename, CtrlHandles *h,
     fprintf(stream, "\n");
     fprintf(stream, "# Attributes:\n");
     fprintf(stream, "\n");
+
+    /*
+     * Note: we only save writable attributes addressable by X screen
+     * (i.e., we don't look at other target types, yet).
+     */
     
-    for (screen = 0; screen < h->num_screens; screen++) {
+    for (screen = 0; screen < h->targets[X_SCREEN_TARGET].n; screen++) {
+
+        t = &h->targets[X_SCREEN_TARGET].t[screen];
 
         /* skip it if we don't have a handle for this screen */
 
-        if (!h->h[screen]) continue;
+        if (!t->h) continue;
 
         /*
          * construct the prefix that will be printed in the config
@@ -275,7 +281,7 @@ int nv_write_config_file(const char *filename, CtrlHandles *h,
 
         if (conf->booleans &
             CONFIG_PROPERTIES_INCLUDE_DISPLAY_NAME_IN_CONFIG_FILE) {
-            prefix = h->screen_names[screen];
+            prefix = t->name;
         } else {
             snprintf(scratch, 4, "%d", screen);
             prefix = scratch;
@@ -301,7 +307,7 @@ int nv_write_config_file(const char *filename, CtrlHandles *h,
             
             if (a->flags & NV_PARSER_TYPE_COLOR_ATTRIBUTE) {
                 float c[3], b[3], g[3];
-                status = NvCtrlGetColorAttributes(h->h[screen], c, b, g);
+                status = NvCtrlGetColorAttributes(t->h, c, b, g);
                 if (status != NvCtrlSuccess) continue;
                 fprintf(stream, "%s%c%s=%f\n",
                         prefix, DISPLAY_NAME_SEPARATOR, a->name,
@@ -313,19 +319,17 @@ int nv_write_config_file(const char *filename, CtrlHandles *h,
                 
                 mask = 1 << bit;
                 
-                if ((h->d[screen] & mask) == 0x0) continue;
+                if ((t->d & mask) == 0x0) continue;
 
-                status =
-                    NvCtrlGetValidDisplayAttributeValues(h->h[screen], mask,
-                                                         a->attr, &valid);
+                status = NvCtrlGetValidDisplayAttributeValues
+                    (t->h, mask, a->attr, &valid);
 
                 if (status != NvCtrlSuccess) goto exit_bit_loop;
                 
                 if ((valid.permissions & ATTRIBUTE_TYPE_WRITE) == 0x0)
                     goto exit_bit_loop;
                 
-                status = NvCtrlGetDisplayAttribute(h->h[screen], mask,
-                                                   a->attr, &val);
+                status = NvCtrlGetDisplayAttribute(t->h, mask, a->attr, &val);
                 
                 if (status != NvCtrlSuccess) goto exit_bit_loop;
                 
@@ -369,28 +373,56 @@ int nv_write_config_file(const char *filename, CtrlHandles *h,
      */
 
     while (p) {
-        
+        char target_str[64];
+
         if (!p->next) {
             p = p->next;
             continue;
         }
 
         tmp = nv_get_attribute_name(p->attr);
-        if (!tmp) continue;
+        if (!tmp) {
+            nv_error_msg("Failure to save unknown attribute %d.", p->attr);
+            p = p->next;
+            continue;
+        }
 
+        /*
+         * if the parsed attribute has a target specification, and a
+         * target type other than an X screen, include a target
+         * specification in what we write to the .rc file.
+         */
+        
+        target_str[0] = '\0';
+        
+        if ((p->flags & NV_PARSER_HAS_TARGET) &&
+            (p->target_type != NV_CTRL_TARGET_TYPE_X_SCREEN)) {
+            
+            int j;
+            
+            /* Find the target name of the target type */
+            for (j = 0; targetTypeTable[j].name; j++) {
+                if (targetTypeTable[j].nvctrl == p->target_type) {
+                    snprintf(target_str, 64, "[%s:%d]",
+                             targetTypeTable[j].parsed_name, p->target_id);
+                    break;
+                }
+            }
+        }
+        
         if (p->display_device_mask) {
-
+            
             tmp_d_str = display_device_mask_to_display_device_name
                 (p->display_device_mask);
-                
-            fprintf(stream, "%s%c%s[%s]=%d\n", p->display,
+            
+            fprintf(stream, "%s%s%c%s[%s]=%d\n", p->display, target_str,
                     DISPLAY_NAME_SEPARATOR, tmp, tmp_d_str, p->val);
             
             free(tmp_d_str);
             
         } else {
                 
-            fprintf(stream, "%s%c%s=%d\n", p->display,
+            fprintf(stream, "%s%s%c%s=%d\n", p->display, target_str,
                     DISPLAY_NAME_SEPARATOR, tmp, p->val);
         }
         
@@ -426,38 +458,47 @@ int nv_write_config_file(const char *filename, CtrlHandles *h,
  */
 
 static ParsedAttributeWrapper *parse_config_file(char *buf, const char *file,
+                                                 const int length,
                                                  ConfigProperties *conf)
 {
-    int line, has_data, len, n, ret;
-    char *cur, *c, *comment, tmp[MAX_CONFIG_FILE_LINE_LEN];
+    int line, has_data, current_tmp_len, len, n, ret;
+    char *cur, *c, *comment, *tmp;
     ParsedAttributeWrapper *w;
     
     cur = buf;
     line = 1;
+    current_tmp_len = 0;
     n = 0;
     w = NULL;
+    tmp = NULL;
 
     while (cur) {
         c = cur;
         comment = NULL;
         has_data = NV_FALSE;;
         
-        while((*c != '\n') && (*c != '\0') && (*c != EOF)) {
+        while (((c - buf) < length) &&
+               (*c != '\n') &&
+               (*c != '\0') &&
+               (*c != EOF)) {
             if (comment) { c++; continue; }
             if (*c == '#') { comment = c; continue; }
             if (!isspace(*c)) has_data = NV_TRUE;
             c++;
         }
-
+        
         if (has_data) {
             if (!comment) comment = c;
             len = comment - cur;
-            if (len >= MAX_CONFIG_FILE_LINE_LEN) {
-                nv_error_msg("Error parsing configuration file '%s' on "
-                             "line %d: line length exceeds maximum "
-                             "length of %d.",
-                             file, line, MAX_CONFIG_FILE_LINE_LEN);
-                goto failed;
+            
+            /* grow the tmp buffer if it's too small */
+            
+            if (len >= current_tmp_len) {
+                current_tmp_len = len + 1;
+                if (tmp) {
+                    free(tmp);
+                }
+                tmp = malloc(sizeof(char) * current_tmp_len);
             }
 
             strncpy (tmp, cur, len);
@@ -484,12 +525,12 @@ static ParsedAttributeWrapper *parse_config_file(char *buf, const char *file,
             }
         }
 
-        if ((*c == '\0') || (*c == EOF)) cur = NULL;
+        if (((c - buf) >= length) || (*c == '\0') || (*c == EOF)) cur = NULL;
         else cur = c + 1;
         
         line++;
     }
-
+    free(tmp);
     /* mark the end of the array */
 
     w = realloc(w, sizeof(ParsedAttributeWrapper) * (n+1));
@@ -499,6 +540,7 @@ static ParsedAttributeWrapper *parse_config_file(char *buf, const char *file,
 
  failed:
     if (w) free(w);
+    free(tmp);
     return NULL;
 
 } /* parse_config_file() */
@@ -567,9 +609,11 @@ static int process_config_file_attributes(const char *file,
                                           "on line %d of configuration file "
                                           "'%s'", w[i].line, file);
         /*
-         * XXX should we fail if processing the attribute failed?  For
-         * now, we'll just keep going through the rest of the config
-         * file.
+         * We do not fail if processing the attribute failed.  If the
+         * GPU or the X config changed (for example stereo is
+         * disabled), some attributes written in the config file may
+         * not be advertised by the the NVCTRL extension (for example
+         * the control to force stereo)
          */
     }
     
@@ -746,7 +790,7 @@ static void write_config_properties(FILE *stream, ConfigProperties *conf, char *
  * structure.
  */
 
-static void init_config_properties(ConfigProperties *conf)
+void init_config_properties(ConfigProperties *conf)
 {
     memset(conf, 0, sizeof(ConfigProperties));
 
