@@ -242,6 +242,8 @@ static void ctk_event_class_init(CtkEventClass *ctk_event_class)
     MAKE_SIGNAL(NV_CTRL_FSAA_APPLICATION_ENHANCED);
     MAKE_SIGNAL(NV_CTRL_FRAMELOCK_SYNC_RATE_4);
     MAKE_SIGNAL(NV_CTRL_GVO_LOCK_OWNER);
+    MAKE_SIGNAL(NV_CTRL_REFRESH_RATE_3);
+    MAKE_SIGNAL(NV_CTRL_ONDEMAND_VBLANK_INTERRUPTS);
 
 #undef MAKE_SIGNAL
     
@@ -252,7 +254,7 @@ static void ctk_event_class_init(CtkEventClass *ctk_event_class)
      * knows about.
      */
 
-#if NV_CTRL_LAST_ATTRIBUTE != NV_CTRL_GVO_LOCK_OWNER
+#if NV_CTRL_LAST_ATTRIBUTE != NV_CTRL_ONDEMAND_VBLANK_INTERRUPTS
 #warning "There are attributes that do not emit signals!"
 #endif
 
